#!/usr/bin/env python3
# csgoMP.  a csgo music player / soundboard for Linux
# Version 0.1.4
# (c) 2021 Adam Feakin csgomp@csparker.uk

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk
import os
import shutil
import subprocess
import sys
import re
from configparser import ConfigParser

# for checking youtube-dl version
import requests

# TODO: maybe make this optional?
from Xlib.display import Display
from Xlib import X
from Xlib.ext import record
from Xlib.protocol import rq
from threading import Thread

# TODO: and this optional too
gi.require_version('Notify', '0.7')
from gi.repository import Notify

# debug
import syslog

disp = None



# Change selection when hotkey is pressed
# TODO: use notify to show popup of what track was selected
def track_selector(reply):
    data = reply.data
    while len(data):
        event, data = rq.EventField(None).parse_binary_value(data,
                                                             disp.display, None, None)

        # print("Keycode pressed: " + str(event.detail))

        # FIXME: Hardcoded keycodes for numpad numbers 0-9 for now.  Need a way to 
        # make this configurable.
        # Change selection on the tree view as if we has clicked it
        if event.detail == 90:
            if len(liststore) > 0:
                Notify.Notification.new("Selected " + liststore[0][3]).show()
            treeview.get_selection().select_path(0)
        elif event.detail == 87:
            if len(liststore) > 1:
                Notify.Notification.new("Selected " + liststore[1][3]).show()
            treeview.get_selection().select_path(1)
        elif event.detail == 88:
            if len(liststore) > 2:
                Notify.Notification.new("Selected " + liststore[2][3]).show()
            treeview.get_selection().select_path(2)
        elif event.detail == 89:
            if len(liststore) > 3:
                Notify.Notification.new("Selected " + liststore[3][3]).show()
            treeview.get_selection().select_path(3)
        elif event.detail == 83:
            if len(liststore) > 4:
                Notify.Notification.new("Selected " + liststore[4][3]).show()
            treeview.get_selection().select_path(4)
        elif event.detail == 84:
            if len(liststore) > 5:
                Notify.Notification.new("Selected " + liststore[5][3]).show()
            treeview.get_selection().select_path(5)
        elif event.detail == 85:
            if len(liststore) > 6:
                Notify.Notification.new("Selected " + liststore[6][3]).show()
            treeview.get_selection().select_path(6)
        elif event.detail == 79:
            if len(liststore) > 7:
                Notify.Notification.new("Selected " + liststore[7][3]).show()
            treeview.get_selection().select_path(7)
        elif event.detail == 80:
            if len(liststore) > 8:
                Notify.Notification.new("Selected " + liststore[8][3]).show()
            treeview.get_selection().select_path(8)
        elif event.detail == 81:
            if len(liststore) > 9:
                Notify.Notification.new("Selected " + liststore[9][3]).show()
            treeview.get_selection().select_path(9)

def latest_youtube_dl():
    # https://yt-dl.org/downloads/latest/ redirects to url with latest version
    redir = requests.get("https://yt-dl.org/downloads/latest/")
    if(redir.history[1]):
        # https://yt-dl.org/downloads/2021.01.24.1/
        url = redir.history[1].url
        latest_version = re.match('.*\/([0-9]{4})\.([0-9]{2})\.([0-9]{2})(\.[0-9])?\/$', url)
        if(latest_version.group(4)):
            patch = int(float(latest_version.group(4))*10)
        else:
            patch = 0
        return(int(latest_version.group(1))*100000 + int(latest_version.group(2))*1000 + int(latest_version.group(3))*10 + patch)
    else:
        return(None)

def my_youtube_dl(path):
    ver = subprocess.check_output(path + " --version", shell=True).rstrip().decode()
    my_version = re.match('^([0-9]{4})\.([0-9]{2})\.([0-9]{2})(\.[0-9])?$', ver)
    if(my_version.group(4)):
        patch = int(float(my_version.group(4))*10)
    else:
        patch = 0
    return(int(my_version.group(1))*100000 + int(my_version.group(2))*1000 + int(my_version.group(3))*10 + patch)


class Handler:
    def onDeleteWindow(self, *args):
        Notify.uninit()
        disp.record_free_context(ctx)

        save_prefs()
        Gtk.main_quit(*args)
        quit_mp(0)

    def test_youtube_dl(self, button):
        youtubedl_path_entry = builder.get_object('entry_youtubedl_path')
        youtube_dl_testpath = youtubedl_path_entry.get_text()
        print("testing path: " + youtube_dl_testpath)
        youtubedl_check = builder.get_object("dialog_prefs_check")
        youtubedl_check_message = builder.get_object("dialog_prefs_check_message")
        # check it's executable
        if(not os.path.isfile(youtube_dl_testpath)):
            print("Error: youtube-dl path doesn't exist")
            youtubedl_check.show()
            youtubedl_check_message.set_text("Couldn't find youtube-dl at the given path")
        else:
            if(not os.access(youtube_dl_testpath, os.X_OK)):
               print("Error: youtube-dl file isn't executable")
               youtubedl_check.show()
               youtubedl_check_message.set_text("Couldn't run youtube-dl, it's not executable")
            else:
                # check the latest version
                latest = latest_youtube_dl()
                mine = my_youtube_dl(youtube_dl_testpath)
                if(not (mine or latest) ):
                    youtubedl_check.show()
                    youtubedl_check_message.set_text("Error checking versions")
                elif(mine < latest):
                    youtubedl_check.show()
                    youtubedl_check_message.set_text("Your youtube-dl is out of date, latest is " + str(latest))
                else:
                    youtubedl_check.show()
                    youtubedl_check_message.set_text("Your youtube-dl seems to be up to date")

    def test_csgo_dir(self, button):
        csgo_dir_entry = builder.get_object("entry_csgo_dir")
        csgo_dir_test = csgo_dir_entry.get_text()
        print("Checking csgo directory: " + csgo_dir_test)
        csgo_dir_check = builder.get_object("dialog_prefs_check")
        csgo_dir_check_message = builder.get_object("dialog_prefs_check_message")
        # check it contains the csgo/cfg directory and steam_appid.txt contains 730
        if(not os.path.isdir(csgo_dir_test)):
            print("Error: csgo_dir doesn't exist")
            csgo_dir_check.show()
            csgo_dir_check_message.set_text("The given CS:GO directory doesn't exist")
            return
        if(not os.path.isdir(csgo_dir_test + "/csgo/cfg/")):
            print("Error: That doesn't look like the csgo directory")
            csgo_dir_check.show()
            csgo_dir_check_message.set_text("That doesn't look like the right CS:GO directory")
            return
        if(os.path.isfile(csgo_dir_test+"/steam_appid.txt")):
            with open(csgo_dir_test+"/steam_appid.txt", 'r') as appid_file:
                appid = appid_file.read()
                if(re.search("730", appid)):
                    print("Looks like csgo")
                else:
                    print("Error: wrong game directory")
                    csgo_dir_check.show()
                    csgo_dir_check_message.set_text("The given directory is for a different Steam game")
                    return
        if(not os.access(csgo_dir_test, os.W_OK)):
            print("Error: can't write to csgo directory")
            csgo_dir_check.show()
            csgo_dir_check_message.set_text("No write access to the CS:GO directory")
            return
        # if we are here it should be ok.
        csgo_dir_check.show()
        csgo_dir_check_message.set_text("CS:GO directory seems to be OK")


    def test_ffmpeg(self, button):
        ffmpeg_path_entry = builder.get_object("entry_ffmpeg_path")
        ffmpeg_testpath = ffmpeg_path_entry.get_text()
        print("Checking ffmpeg version: " + ffmpeg_testpath)
        ffmpeg_check = builder.get_object("dialog_prefs_check")
        ffmpeg_check_message = builder.get_object("dialog_prefs_check_message")
        # check it exists, is executable and print version
        if(not (os.path.isfile(ffmpeg_testpath) and os.access(ffmpeg_testpath, os.X_OK))):
           ffmpeg_check.show()
           ffmpeg_check_message.set_text("Given ffmpeg path doesn't exist or is not executable")
           return
        ffmpeg_output = subprocess.check_output(ffmpeg_testpath + " -version", shell=True).rstrip().decode()
        version = re.match('^ffmpeg version ([0-9]+\.[0-9]+\.[0-9]+) Copyright', ffmpeg_output)
        ffmpeg_check.show()
        if(version):
            ffmpeg_check_message.set_text("ffmpeg version " + version.group(1) + " found")
        else:
            ffmpeg_check_message.set_text("ffmpeg found, but unknown version")


    def prefs_check_ok_cb(self, button):
        builder.get_object("dialog_prefs_check").hide()

    def show_prefs_cb(self, button):
        # some options
        # csgo install path
        # keybinding for play/stop
        # choose max audio file length, cut if too long (ffmpeg -t option)
        print("Showing config window...")
        push_prefs()
        prefswin.show()

    def save_cfg_cb(self, button):
        print("Save cfg...")
        global ffmpeg_path, csgo_directory, files_count, \
                youtube_dl_path
        csgo_cfg_path = csgo_directory + "/csgo/cfg"
        cfg_filename = csgo_cfg_path + "/csgomp.cfg"

        # delete exising file
        if os.path.isfile(cfg_filename) == True:
            print("Deleting old cfg")
            os.remove(cfg_filename)

        # make a new one
        with open(cfg_filename, "w") as f:
            f.write("echo *** Loading csgomp config ***\n");
            f.write("echo Push to play key: ' \n");
            f.write("echo Toggle playback key: / \n");
            f.write("alias +csgomp_play csgomp_start\n")
            f.write("alias -csgomp_play csgomp_stop\n")
            f.write("alias csgomp_start \"voice_inputfromfile 1; voice_loopback 1; +voicerecord\"\n")
            f.write("alias csgomp_stop \"-voicerecord; voice_loopback 0; voice_inputfromfile 0\"\n")
            # TODO: add a pref to set this bind key, and the continuous play key too
            f.write("bind ' +csgomp_play\n")
            f.write("alias csgomp_toggle_on \"csgomp_start; bind / "
                    "csgomp_toggle_off\"\n")
            f.write("alias csgomp_toggle_off \"csgomp_stop; bind / "
                    "csgomp_toggle_on\"\n")
            f.write("bind / csgomp_toggle_on\n")
            f.close()

            #TODO: use pyinotify/watchdog or similar to check a request file for
            #the audio track. Just poll every 3 secs if neither available

        # TODO: add a popup to tell user to "exec csgomp" in console
        #show_cfg_load_popup()

    def prefs_ok_cb(self, button):
        print("Saving settings...")
        pull_prefs()
        # TODO: Check ffmpeg, csgo and youtube-dl paths exist and are executable (and dir for csgo)
        save_prefs()
        prefswin.hide()

    def prefs_cancel_cb(self, button):
        print("Not saving settings")
        prefswin.hide()

    def prefs_close_cb(self, button):
        print("WM killed window")
        #prefswin.hide()
        #prefswin.destroy()

    def add_file_cb(self, button):
        print("Show filesel...")
        # don't fetch and destroy here, just use run/show and hide
        resp = fileswin.run()
        if resp == 1:
            print("OK")
            filename = fileswin.get_filename()
            print("Filename is " + filename)
            alias = builder.get_object("entry_filesel_alias").get_text()
            print("Alias is " + alias)
            #FIXME: ask for start pos, end pos/duration and alias, and autoplay
            file_start = builder.get_object("entry_filesel_start_time").get_text()
            file_end = builder.get_object("entry_filesel_end_time").get_text()

            # no duration set yet
            duration = -1

            if(not file_start and not file_end):
                print("encoding whole file")
                start_pos = 0;
                end_pos = 0;
                duration = 0
            else:
                if(file_start):
                    print("starting from " + file_start)
                    start_pos = parse_time(file_start)
                if(file_end):
                    print("stopping at " + file_end)
                    end_pos = parse_time(file_end)
                duration = end_pos - start_pos
                print("duration is " + str(duration))
                print("ffmpeg opt: -ss " + str(start_pos) + " -t " + str(duration))

            if duration < 0:
                print("Error: duration can't be less than 0")
                duration = 0

            keybinding = str(len(liststore))
            add_to_list(filename, str(start_pos) , str(duration), alias, keybinding, False)

            # debug: print what is in the list
            treeiter = liststore.get_iter_first()
            while treeiter != None:
                print(str(liststore[treeiter][0]))
                treeiter = liststore.iter_next(treeiter)

        elif resp == 2:
            print("CANCEL")

        fileswin.hide()

    def youtube_ok_cb(self, button):
        print("Add from youtube")
        # get url, start and end (if entered)
        yt_url = builder.get_object("entry_youtube_url").get_text()
        yt_start = builder.get_object("entry_youtube_start_time").get_text()
        yt_end = builder.get_object("entry_youtube_end_time").get_text()
        yt_alias = builder.get_object("entry_youtube_alias").get_text()
        yt_volume = builder.get_object("scale_youtube_volume").get_value()

        # TODO: write parse_time function to check it really is a time
        # and convert to seconds from MM:SS.
        print("URL: " + yt_url + " start: " + yt_start + " end: " + yt_end + 
              " volume: " + str(yt_volume))

        # no duration set yet
        duration = -1

        end_pos = 0
        start_pos = 0

        if(not yt_start and not yt_end):
            print("encoding whole file")
            start_pos = 0;
            end_pos = 0;
        else:
            if(yt_start):
                print("starting from " + yt_start)
                start_pos = parse_time(yt_start)
            if(yt_end):
                print("stopping at " + yt_end)
                end_pos = parse_time(yt_end)
            duration = end_pos - start_pos
            print("duration is " + str(duration))
            print("ffmpeg opt: -ss " + str(start_pos) + " -t " + str(duration))

        if duration < 0:
            print("Error: duration can't be less than 0")
            duration = 0

        # TODO: write to /tmp, make it work with cache_name somehow
        grab_youtube(yt_url, start_pos, duration, yt_volume, yt_alias)

        youtubewin.hide()

    def youtube_cancel_cb(self, button):
        print("Cancel youtube add")
        youtubewin.hide()

    def add_youtube_cb(self, button):
        if(youtube_dl_path):
            print("Show youtube dialog...")
            youtubewin.show()

            # set up the volume selector
            scale_youtube_volume = builder.get_object("scale_youtube_volume")
            scale_youtube_volume.set_range(0, 100)
            scale_youtube_volume.set_increments(1, 10)
            scale_youtube_volume.set_digits(0)
            scale_youtube_volume.set_value(100)

        else:
            print("youtube-dl not found")
            builder.get_object("dialog_noyoutubedl").show()

    def filesel_open_clicked_cb(self, button):
        print("Open")
        fileswin.destroy()

    def filesel_cancel_clicked_cb(self, button):
        print("Cancel")
        fileswin.destroy()

    def noytdl_ok_cb(self, button):
        builder.get_object("dialog_noyoutubedl").hide()

    def nocsgodir_ok_cb(self, button):
        builder.get_object("dialog_nocsgodir").hide()

    def youtubedlerror_ok_cb(self, button):
        builder.get_object("dialog_youtubedlerror").hide()

# called when a file is selected from the list
# convert the file to wav for use in game
# TODO: add right click callback for changing alias and keybind
# TODO: when a file is added or keybind change highlight the "write csgo cfg"
# button so the user knows they need to click it
    def selection_changed_cb(self, selection):
        model, treeiter = selection.get_selected()
        global ffmpeg_path, csgo_directory, files_count, \
                youtube_dl_path
        if treeiter != None:
            # update voice_input.wav to selected file
            wav_file = csgo_directory + "/voice_input.wav"

            audio_len_flag = ""

            # TODO: check file still exists
            convertme = model[treeiter][0]
            start_pos = int(model[treeiter][1])
            duration = int(model[treeiter][2])
            alias = model[treeiter][3]


            # TODO: clear cache button

            # check for file in cache using convertme + start_pos + length as
            # name. eg jamesbond.mp3_3_10.wav for jamesbond.mp3 starting from 3
            # seconds in and playing for 10 secs.  jamesbond.mp3_0_0.wav for full
            # track.
            # TODO: remove the max_audio_length option and save a starttime and
            # duration with every filename
            cache_name, in_cache = check_cache(convertme, start_pos, duration)
            if in_cache:
                # use cached version
                print("WAV in cache, copying")
                if os.path.isfile(wav_file):
                    os.remove(wav_file)
                shutil.copy(cache_name, wav_file)
            else:
                print("Not cached, converting...")
                print("convertme: " + convertme + " start_pos: " +
                      str(start_pos) +
                      "duration:" + str(duration));
                if(start_pos == 0 and duration == 0):
                    # whole song
                    print("Encoding whole song...");
                    audio_len_flag = " "
                elif(start_pos > 0 and duration > 0):
                    # startpos and duration set
                    audio_len_flag = " -ss " + str(start_pos) + " -t " + str(duration)
                elif(start_pos > 0):
                    # just a start position
                    audio_len_flag = " -ss " + str(start_pos)
                elif(duration > 0):
                    # just a duration
                    audio_len_flag = " -t " + str(duration)

                # TODO: check first if this is a youtube link that needs
                # downloading.  should only happen if cache has been cleared
                if(convertme[0] == 'h'):
                    # TODO: add volume to list and cache
                    grab_youtube(convertme, start_pos, duration, 100.0, alias)
                else:
                    # needs -flags +bitexact so ffmpeg doesnt add metadata
                    # for ffmpeg 4 this is "-bitexact", which doesn't work in
                    # ffmpeg 3
                    # also seems to need -map_metadata -1
                    if ffmpeg_version == 3:
                        ffmpeg_opts = " -vn -map_metadata -1 -acodec pcm_s16le -ar 22050 -ac 1 -f wav -flags +bitexact"
                    else:
                        ffmpeg_opts = " -vn -map_metadata -1 -acodec pcm_s16le -ar 22050 -ac 1 -f wav -bitexact"

                    # use -y option to force overwrite just incase
                    ffmpeg_cmd = ffmpeg_path + " -loglevel panic -y -i '" +convertme + "' " + audio_len_flag + ffmpeg_opts + " '" + cache_name + "'"

                    print("cmd: " + ffmpeg_cmd)

                    #FIXME: why doesnt call work?
                    #call([ffmpeg_path, ffmpeg_cmd])
                    os.system(ffmpeg_cmd)

                    # delete any existing file
                    if os.path.isfile(wav_file):
                        os.remove(wav_file)
                    shutil.copy(cache_name, wav_file)


    # called when clicking a list item (with any button)
    def list_item_click_cb(self, treeview, event):
        # check if this is a right click
        if event.button == 3:
            # get path
            global pthinfo
            pthinfo = treeview.get_path_at_pos(int(event.x), int(event.y))
            if pthinfo is not None:
                path, col, cellx, celly = pthinfo
                # get the tree view model and print the filename at 
                # the path at pointer position
                model = treeview.get_model()
                print("Right click " + model[path][0])
                popup = builder.get_object("menu_popup")
                # XXX: save the pthinfo as a global so it can be used
                # by the callback functions.  I couldn't find another
                # way to get the data to the callbacks
                popup.popup(None, None, None, None, event.button, event.time)

    # context menu callbacks
    # use path from pthinfo to get the data to work on
    # and insert it into the dialog boxes.
    # dialog box save button updates the treeview data
    def popup_edit_cb(self, menuitem):
        print(pthinfo)
        if pthinfo is not None:
            path, col, cellx, celly = pthinfo
            model = treeview.get_model()
            print("Going to edit: " + model[path][0])

    def popup_remove_cb(self, menuitem):
        print(pthinfo)
        if pthinfo is not None:
            path, col, cellx, celly = pthinfo
            model = treeview.get_model()
            remove_filename = model[path][0];
            print("Going to remove: " + remove_filename)
            for row in liststore:
                if row[0] == remove_filename:
                    # TODO: remove from cache too
                    cache_name, in_cache = check_cache(row[0], row[1], row[2])
                    if in_cache == True:
                        print("Deleting: " + cache_name)
                        os.remove(cache_name)

                    # remove from list last so we can still look up the name
                    # and stuff
                    liststore.remove(row.iter)

                    # redo keybindings, I guess this is a hack since you can't
                    # choose your own
                    redo_keybindings()
                    break




# end: Handler

def redo_keybindings():
    i = 0
    for row in liststore:
        row[4] = str(i)
        i = i+1


# check cache first
# use youtube-dl to get a file from youtube and save in /tmp
# 
def grab_youtube(url, start_pos:int , duration:int , volume, alias):
    # TODO: use volume in cache name too
    cache_name, in_cache = check_cache(url, start_pos, duration)
    wav_file = csgo_directory + "/voice_input.wav"

    if not os.path.isdir(csgo_directory):
        # show error window
        builder.get_object("dialog_nocsgodir").show()

        # exit
        return

    if in_cache:
        # use cached version
        print("WAV in cache, copying")
        if os.path.isfile(wav_file):
            os.remove(wav_file)
        shutil.copy(cache_name, wav_file)
        # add it to the file list

        # FIXME: ask for alias, autoplay
        keybinding = str(len(liststore))
        add_to_list(url, str(start_pos), str(duration), alias, keybinding,
                         autoplay)
    else:
        print("Not cached, downloading and converting...")
        print("target name " + cache_name)

        # download to /tmp using youtube-dl
        # first get output filename
        youtube_dl_opts = "-o '/tmp/csgompyttmp.%(ext)s' '" + url + "' -f mp4 --get-filename"
        youtube_dl_cmd = youtube_dl_path + " " + youtube_dl_opts
        print("youtube-dl cmd: " + youtube_dl_cmd)
        try:
            yt_dl_filename = subprocess.check_output(youtube_dl_cmd,
                                                     shell=True).rstrip().decode()
            print("yt_dl_filename: " + yt_dl_filename)

        except subprocess.CalledProcessError:
            builder.get_object("dialog_youtubedlerror").show()
            return

        print("Got filename:[" + yt_dl_filename + "]")
        # make sure it doesn't already exist
        if os.path.isfile(yt_dl_filename):
            os.remove(yt_dl_filename)

        # then actually download
        youtube_dl_opts = "-o '/tmp/csgompyttmp.%(ext)s' '" + url + "' -f mp4"
        youtube_dl_cmd = youtube_dl_path + " " + youtube_dl_opts
        print("youtube-dl cmd: " + youtube_dl_cmd)

        try:
            subprocess.call(youtube_dl_cmd, shell=True)
        except subprocess.CalledProcessError:
            builder.get_object("dialog_youtubedlerror").show()
            return

        if not os.path.isfile(yt_dl_filename):
            print("Downloaded file not found")

        # then convert to wav with ffmpeg with start_pos and duration
        if ffmpeg_version == 3:
            ffmpeg_opts = " -vn -map_metadata -1 -acodec pcm_s16le -ar 22050 -ac 1 -f wav -flags +bitexact "
        else:
            ffmpeg_opts = " -vn -map_metadata -1 -acodec pcm_s16le -ar 22050 -ac 1 -f wav -bitexact "

        if(start_pos == 0 and duration == 0):
            # whole song
            audio_len_flaga = " "
            audio_len_flagb = " "
        elif(start_pos > 0 and duration > 0):
            # startpos and duration set
            audio_len_flaga = " -ss " + str(start_pos)
            audio_len_flagb = " -t " + str(duration)
        elif(start_pos > 0):
            # just a start position
            audio_len_flaga = " -ss " + str(start_pos)
            audio_len_flagb = " "
        elif(duration > 0):
            # just a duration
            audio_len_flaga = " "
            audio_len_flagb = " -t " + str(duration)

        if(volume != 100):
            volume_1 = volume / 100
            volume_flags = " -filter:a volume='" + str(volume_1) + "' "
            print("volume_flags: [" + volume_flags + "]")
        else:
            volume_flags = " "

        # use -y option to force overwrite just incase
        ffmpeg_cmd = ffmpeg_path + " -y " + audio_len_flaga + " -i '" + str(yt_dl_filename) + "' " + audio_len_flagb + volume_flags + ffmpeg_opts + " '" + cache_name + "'"
        print("cmd: " + ffmpeg_cmd)
        subprocess.call(ffmpeg_cmd, shell=True)

        # copy it to wav_file
        if os.path.isfile(wav_file):
            os.remove(wav_file)
        shutil.copy(cache_name, wav_file)

        # add it to the file list
        # FIXME: ask for alias, autoplay on keybind
        keybinding = str(len(liststore))
        add_to_list(url, str(start_pos), str(duration), alias, keybinding,
                     autoplay)

        # and delete the tmp file
        os.remove(yt_dl_filename)

# try to convert time format into seconds
# formats hh:mm:ss, mm:ss, ss
def parse_time(time_str):

    time_str = re.sub("[^0-9:.]", "", time_str);

    if time_str.count(':') == 2:
        print("hh:mm:ss")
        h, m, s = time_str.split(':')
        return (int(h) * 3600) + (int(m) * 60) + int(s)
    elif time_str.count(':') == 1:
        print("mm:ss")
        m, s = time_str.split(':')
        return (int(m) * 60) + int(s)
    elif time_str.count(':') == 0:
        print("ss")
        return int(time_str)
    else:
        print("Time format error")


# check for file in the cache, return the cache file name and
# true or false
def check_cache(convertme, start_time, audio_len):
    cache_name = ""
    in_cache = False

    # check cache directory exists, create it if not
    if not os.path.isdir(cache_directory):
        print("Creating " + cache_directory)
        os.mkdir(cache_directory)

    # create the filename
    # is this a youtube url?
    if convertme.startswith("http"):
        cache_name = cache_directory + "/" + convertme.replace("/","-") + "_" + str(start_time) + "_" + str(audio_len) + ".wav"
    else:
        cache_name = cache_directory + "/" + os.path.basename(convertme) + "_" + str(start_time) + "_" + str(audio_len) + ".wav"
    print("cache_name is " + cache_name)

    # check file exists
    in_cache = os.path.isfile(cache_name)

    return cache_name, in_cache

def push_prefs():
    # update prefs window with current values
    global ffmpeg_path, csgo_directory, files_count,\
            youtube_dl_path
    builder.get_object("entry_ffmpeg_path").set_text(ffmpeg_path)
    builder.get_object("entry_csgo_dir").set_text(csgo_directory)
    if(youtube_dl_path):
        builder.get_object("entry_youtubedl_path").set_text(youtube_dl_path)

def pull_prefs():
    # get values from prefs boxes into variables
    global ffmpeg_path, csgo_directory, files_count,\
            youtube_dl_path
    ffmpeg_path = builder.get_object("entry_ffmpeg_path").get_text()
    csgo_directory = builder.get_object("entry_csgo_dir").get_text()
    youtube_dl_path = builder.get_object("entry_youtubedl_path").get_text()

    print(ffmpeg_path)


# add to the list if filename + duration + start_pos aren't already there
def add_to_list(url, start_pos, duration, alias, keybinding, autoplay):
    for row in liststore:
        if row[0] == url and row[1] == start_pos and row[2] == duration:
            return
    liststore.append([url, start_pos, duration, alias, keybinding, autoplay])


def load_prefs():
    global ffmpeg_path, csgo_directory, files_count, youtube_dl_path
    if os.path.isfile(os.path.dirname(os.path.realpath(__file__)) + "/config.ini") == False:
        print("No settings found")
        print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
        print("!! If this is your first time running then you may need to run csgomp !!")
        print("!! without any arguments and set the correct paths in the preferences !!")
        print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
        return

    # load prefs
    config = ConfigParser()
    config.read(os.path.dirname(os.path.realpath(__file__)) + "/config.ini")

    # ffmpeg_path
    ffmpeg_path = config.get("main", "ffmpeg_path")

    # csgo_directory
    csgo_directory = config.get("main", "csgo_directory")

    # youtube-dl
    if config.has_option("main", "youtube_dl_path"):
        youtube_dl_path = config.get("main", "youtube_dl_path")

    # files_count
    files_count = config.getint("main", "files_count")

    # window size
    if config.has_option("main", "window_width"):
        w = config.getint("main", "window_width")
        h = config.getint("main", "window_height")
        mainwin.resize(w, h)

    # load files section
    i = 0
    while i < files_count:
        filename = config.get("files", "file"+str(i))
        start = config.get("files", "start"+str(i))
        dur = config.get("files", "dur"+str(i))
        alias = config.get("files", "alias"+str(i))
        keybinding = config.get("files", "keybinding"+str(i))
        autoplay = config.getboolean("files", "autoplay"+str(i))
        print("Found " + filename)
        keybinding = str(len(liststore))
        add_to_list(filename, start, dur, alias, keybinding, autoplay)
        i+=1

def save_prefs():
    # save prefs
    global ffmpeg_path, csgo_directory, files_count, \
            youtube_dl_path
    print("Saving...")
    config = ConfigParser()
    config.read(os.path.dirname(os.path.realpath(__file__)) + "/config.ini")

    if config.has_section("main") == False:
        config.add_section("main")

    # ffmpeg_path
    config.set("main", "ffmpeg_path", ffmpeg_path)
    print(ffmpeg_path)

    # csgo_directory
    config.set("main", "csgo_directory", csgo_directory)
    print(csgo_directory)

    # youtube-dl
    if(youtube_dl_path):
        config.set("main", "youtube_dl_path", youtube_dl_path)
        print(youtube_dl_path)

    # window size
    (w, h) = mainwin.get_size()
    print("w: " + str(w) + " h: " + str(h))
    config.set("main", "window_width", str(w))
    config.set("main", "window_height", str(h))


    # save files section
    if config.has_section("files") == False:
        config.add_section("files")

    treeiter = liststore.get_iter_first()
    files_count = 0
    while treeiter != None:
        filename = str(liststore[treeiter][0])
        config.set("files", "file"+str(files_count), filename)
        config.set("files", "start"+str(files_count),
                   str(liststore[treeiter][1]))
        config.set("files", "dur"+str(files_count),
                   str(liststore[treeiter][2]))
        config.set("files", "alias"+str(files_count),
                   str(liststore[treeiter][3]))
        config.set("files", "keybinding"+str(files_count),
                   str(liststore[treeiter][4]))
        config.set("files", "autoplay"+str(files_count),
                   str(liststore[treeiter][5]))
        treeiter = liststore.iter_next(treeiter)
        files_count+=1

    # files_count
    config.set("main", "files_count", str(files_count))
    print(str(files_count))


    # TODO: save to ~/.config/csgomp.ini
    with open(os.path.dirname(os.path.realpath(__file__)) + "/config.ini", "w") as f:
        config.write(f)

# function to find executables, taken from
# http://stackoverflow.com/questions/377017/test-if-executable-exists-in-python
def find_exe(program):
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return(program)
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return(exe_file)
    return None


def quit_mp(r: int):
    sys.stdout.close()
    sys.stderr.close()
    os._exit(r)

# Main
# Debug stuff
pid = os.getpid()
STDOUT= '/tmp/csgompout-' + str(pid) + '.txt'
STDERR= '/tmp/csgomperr-' + str(pid) + '.txt'
#print("Redirecting to ", STDOUT, " and ", STDERR)
#sys.stdout = open(STDOUT, 'w')
#sys.stderr = open(STDERR, 'w')

# load the glade file
builder = Gtk.Builder()
builder.add_from_file(os.path.dirname(os.path.realpath(__file__)) + "/MainWindow.glade")
builder.connect_signals(Handler())

# and fetch the windows
mainwin = builder.get_object("window1")
mainwin.set_icon_from_file(os.path.dirname(os.path.realpath(__file__)) + "/icon.png")
mainwin.show_all()

fileswin = builder.get_object("filechooserdialog1")
prefswin = builder.get_object("prefs")
youtubewin = builder.get_object("dialog_add_youtube")

# enable hotkey watcher
disp = Display()
# root = disp.screen().root
# TODO: Can this only watch for keypresses that are actually configured instead
# of everying?
ctx = disp.record_create_context(0,
                                 [record.AllClients],
                                 [{
                                     'core_requests': (0, 0),
                                     'core_replies': (0, 0),
                                     'ext_requests': (0, 0, 0, 0),
                                     'ext_replies': (0, 0, 0, 0),
                                     'delivered_events': (0, 0),
                                     'device_events': (X.KeyPress, X.KeyPress),
                                     'errors': (0, 0),
                                     'client_started': False,
                                     'client_died': False,
                                 }])
print("hi1")
Thread(target=disp.record_enable_context, args=(ctx, track_selector)).start()
print("hi2")


# TODO: add a keybinding column too
# create a liststore for the files list
liststore = Gtk.ListStore(str, str, str, str, str, bool)

treeview = builder.get_object("treeview1")
treeview.set_model(liststore)
# and put a heading
treeview.append_column(Gtk.TreeViewColumn("Filename", Gtk.CellRendererText(), text=0))
treeview.append_column(Gtk.TreeViewColumn("Start", Gtk.CellRendererText(), text=1))
treeview.append_column(Gtk.TreeViewColumn("Duration", Gtk.CellRendererText(), text=2))
treeview.append_column(Gtk.TreeViewColumn("Alias", Gtk.CellRendererText(), text=3))
treeview.append_column(Gtk.TreeViewColumn("Key", Gtk.CellRendererText(), text=4))
treeview.append_column(Gtk.TreeViewColumn("Autoplay", Gtk.CellRendererText(), text=5))

# set some default values
ffmpeg_path = find_exe("ffmpeg")

csgo_directory = os.path.expanduser("~/.local/share/Steam/steamapps/common/Counter-Strike Global Offensive/")
# TODO: try to check csgo directory is correct
files_count = 0
youtube_dl_path = find_exe("youtube-dl")
cache_directory = os.path.expanduser("~/.cache/csgomp")
keybinding = None
autoplay = False

# load prefs
load_prefs()

if ffmpeg_path:
    # check ffmpeg version
    ffmpeg_output = subprocess.check_output([ffmpeg_path,'-version'])
    ffmpeg_match = re.match('^ffmpeg version (\d+)\.(\d+)\.?(.+?)? Copyright.*',
                            ffmpeg_output.decode())
    if ffmpeg_match is None:
        print("Couldn't detect ffmpeg version, assuming 4")
        ffmpeg_version = 4
    else:
        print("Found ffmpeg version " + ffmpeg_match.group(1))
        ffmpeg_version = int(ffmpeg_match.group(1))

# enable notifications
Notify.init("csgoMP")

# If called with command line arguments (except -w) assume the command line is the search
# term. grab_youtube using alias of cmd and url of ytsearch:args. last option
# is a time start point if it starts with +
if len(sys.argv) > 1:
    if sys.argv[1] == '-w':
        print("Starting webserver mode...")
        # initial webserver code stolen from
        # https://gist.github.com/mdonkers/63e115cc0c79b4f6b8b3a6b797e485c7
        try:
            import pyautogui
            from http.server import BaseHTTPRequestHandler, ThreadingHTTPServer
            import urllib.parse

            class req_handler(BaseHTTPRequestHandler):
                global webstatus
                webstatus = "Starting..."

                def _set_response(self, refresh=False, viewport=False):
                    self.send_response(200)
                    self.send_header('Content-type', 'text/html')
                    if refresh:
                        self.send_header('Refresh', "3; url="+refresh)
                    self.end_headers()
                    if viewport:
                        self.wfile.write("<meta name=\"viewport\" content=\"width=device-width, initial-scale=1\">".encode('utf-8'))

                def do_GET(self):
                    global webstatus
                    parsed_url = urllib.parse.urlparse(self.path)
                    queries = urllib.parse.parse_qs(parsed_url.query)

                    #self.wfile.write(("GET request " + self.path + "<br />\n").encode('utf-8'))

                    if(parsed_url.path == '/'):
                        self._set_response(False, True)
                        self.wfile.write("This is the main page<br />"
                                         "Song: <input type=text name=title id=title><br />"
                                         "Start time: <input type=text name=time id=time><br /><button onclick=searchSong() type=button>Search</button><br />"
                                         "<form method=get action=/toggle><input type=submit name=toggle_mic value=ToggleMic></form><br />"
                                         "<div id=status>"
                                         "<script>"
                                         "function searchSong() { var xmlHttp = new XMLHttpRequest();"
                                         "var title = document.getElementById('title').value;"
                                         "var time = document.getElementById('time').value;"
                                         "xmlHttp.open('GET', '/search'+'?title='+encodeURIComponent(title)+'&time='+encodeURIComponent(time), true);"
                                         "xmlHttp.send(null);};"
                                         "window.onload = updateStatus();"
                                         "function updateStatus() { var xmlHttp = new XMLHttpRequest();"
                                         "xmlHttp.open('GET', '/status', false);xmlHttp.send(null);"
                                         "document.getElementById('status').innerHTML = xmlHttp.responseText;"
                                         "setTimeout(updateStatus, 5000);}</script>"
                                         "".encode('utf-8'))
                    if(parsed_url.path == '/search'):
                        # TODO: check if a url was pasted, then no need to search
                        self._set_response('/')
                        query = urllib.parse.unquote_plus(self.path[8:])
                        if 'title' in queries:
                            search_term = "ytsearch:" + queries['title'][0]
                        else:
                            webstatus = "Error, no song title found..."
                            return
                        if 'time' in queries:
                            search_time = queries['time'][0]
                        else:
                            search_time = 0

                        print("Got search_term ", search_term, " and time ", search_time)
                        self.wfile.write(("Searching for: " + self.path[8:]).encode('utf-8'))
                        webstatus = "Searching for " + " " + search_term + "(" + str(search_time) + ")" + "<br />If this gets stuck for over 1 minute maybe youtube blocked or throttled the grab"
                        grab_youtube(search_term, parse_time(str(search_time)), 0, 10.0, "web")
                        webstatus = "Current song: " + search_term # + " Duration: " + yt_duration
                    if(parsed_url.path == '/status'):
                        self._set_response('/status')
                        self.wfile.write(webstatus.encode('utf-8'))
                    if(parsed_url.path == '/toggle'):
                        self._set_response('/')
                        self.wfile.write("Using pyautogui to toggle playback button /".encode('utf-8'))
                        pyautogui.press('/')
                    if(parsed_url.path == '/execcsgomp'):
                        # TODO
                        # pyautogui open console and exec csgomp.cfg. Could also just manually enter all the
                        # commands instead of calling the cfg file.
                        print("`; exec csgomp; `")
                    if(parsed_url.path == '/shutdown'):
                        self._set_response()
                        self.server.server_close()
                        quit_mp(0)


                def do_POST(self):
                    self._set_response()
                    self.wfile.write("POST request .... ")

            def start_webserver(server_class=ThreadingHTTPServer, handler_class=req_handler, port=1337):
                server_address = ('', port)
                httpd = server_class(server_address, handler_class)
                print("Serving...")
                try:
                    httpd.serve_forever()
                except KeyboardInterrupt:
                    pass
                httpd.server_close()
        except ImportError as e:
            print(e)
            print("PyAutoGUI (and http.server) must be installed for webserver")
            quit_mp(1)
        # loop until killed
        start_webserver()
        quit_mp(0)
    else:
        search_term = "ytsearch:" + " ".join(sys.argv[1:])
        search_time = 0
        if len(sys.argv) > 2 and sys.argv[-1][0] == '+':
            search_time = parse_time(sys.argv[-1])
            search_term = "ytsearch:" + " ".join(sys.argv[1:-1])

        Notify.Notification.new("Searching for " + " " + search_term +
                                "(" + str(search_time) + ")").show()

        # check that the duration isn't over 10m, TODO: this can be a cfg option
        try:
            youtube_dl_opts = "--get-duration \"" + search_term + "\""
            youtube_dl_cmd = youtube_dl_path + " " + youtube_dl_opts
            yt_duration = subprocess.check_output(youtube_dl_cmd,
                                                     shell=True).rstrip().decode()
            if parse_time(yt_duration) > 600:
                Notify.Notification.new("Duration is too long " + yt_duration).show();
                quit_mp(1)

        except subprocess.CalledProcessError:
            Notify.Notification.new("Error running youtube-dl").show();
            quit_mp(1)

        # get it, 10% volume TODO: add option for command line default volume
        grab_youtube(search_term, search_time, 0, 10.0, "cmd")
        Notify.Notification.new("Search completed, Duration " + yt_duration).show()
        # exit without showing gui
        quit_mp(0)

# start the GTK UI if there were no args
Gtk.main()

